@@ -14,7 +14,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-from positional import positional
+import six.moves.urllib as urllib
 
 from keystoneclient import base
 from keystoneclient import exceptions
@@ -37,7 +37,6 @@ class Project(base.Resource):
 
     """
 
-    @positional(enforcement=positional.WARN)
     def update(self, name=None, description=None, enabled=None):
         kwargs = {
             'name': name if name is not None else self.name,
@@ -55,6 +54,24 @@ class Project(base.Resource):
 
         return retval
 
+    def add_tag(self, tag):
+        self.manager.add_tag(self, tag)
+
+    def update_tags(self, tags):
+        return self.manager.update_tags(self, tags)
+
+    def delete_tag(self, tag):
+        self.manager.delete_tag(self, tag)
+
+    def delete_all_tags(self):
+        return self.manager.update_tags(self, [])
+
+    def list_tags(self):
+        return self.manager.list_tags(self)
+
+    def check_tag(self, tag):
+        return self.manager.check_tag(self, tag)
+
 
 class ProjectManager(base.CrudManager):
     """Manager class for manipulating Identity projects."""
@@ -63,7 +80,6 @@ class ProjectManager(base.CrudManager):
     collection_key = 'projects'
     key = 'project'
 
-    @positional(3, enforcement=positional.WARN)
     def create(self, name, domain, description=None,
                enabled=True, parent=None, **kwargs):
         """Create a project.
@@ -96,7 +112,6 @@ class ProjectManager(base.CrudManager):
             enabled=enabled,
             **kwargs)
 
-    @positional(enforcement=positional.WARN)
     def list(self, domain=None, user=None, **kwargs):
         """List projects.
 
@@ -106,17 +121,24 @@ class ProjectManager(base.CrudManager):
                      assignments on.
         :type user: str or :class:`keystoneclient.v3.users.User`
         :param kwargs: any other attribute provided will filter projects on.
+                       Project tags filter keyword: ``tags``, ``tags_any``,
+                       ``not_tags``, and ``not_tags_any``. tag attribute type
+                       string. Pass in a comma separated string to filter
+                       with multiple tags.
 
         :returns: a list of projects.
         :rtype: list of :class:`keystoneclient.v3.projects.Project`
 
         """
         base_url = '/users/%s' % base.getid(user) if user else None
-        return super(ProjectManager, self).list(
+        projects = super(ProjectManager, self).list(
             base_url=base_url,
             domain_id=base.getid(domain),
             fallback_to_auth=True,
             **kwargs)
+        for p in projects:
+            p.tags = self._encode_tags(getattr(p, 'tags', []))
+        return projects
 
     def _check_not_parents_as_ids_and_parents_as_list(self, parents_as_ids,
                                                       parents_as_list):
@@ -132,7 +154,6 @@ class ProjectManager(base.CrudManager):
                     'parameters, not both')
             raise exceptions.ValidationError(msg)
 
-    @positional()
     def get(self, project, subtree_as_list=False, parents_as_list=False,
             subtree_as_ids=False, parents_as_ids=False):
         """Retrieve a project.
@@ -180,9 +201,15 @@ class ProjectManager(base.CrudManager):
         query = self.build_key_only_query(query_params)
         dict_args = {'project_id': base.getid(project)}
         url = self.build_url(dict_args_in_out=dict_args)
-        return self._get(url + query, self.key)
+        p = self._get(url + query, self.key)
+        p.tags = self._encode_tags(getattr(p, 'tags', []))
+        return p
+
+    def find(self, **kwargs):
+        p = super(ProjectManager, self).find(**kwargs)
+        p.tags = self._encode_tags(getattr(p, 'tags', []))
+        return p
 
-    @positional(enforcement=positional.WARN)
     def update(self, project, name=None, domain=None, description=None,
                enabled=None, **kwargs):
         """Update a project.
@@ -220,3 +247,82 @@ class ProjectManager(base.CrudManager):
         """
         return super(ProjectManager, self).delete(
             project_id=base.getid(project))
+
+    def _encode_tags(self, tags):
+        """Encode tags to non-unicode string in python2.
+
+        :param tags: list of unicode tags
+
+        :returns: List of strings
+        """
+        return [str(t) for t in tags]
+
+    def add_tag(self, project, tag):
+        """Add a tag to a project.
+
+        :param project: project to add a tag to.
+        :param tag: str name of tag.
+
+        """
+        url = "/projects/%s/tags/%s" % (base.getid(project),
+                                        urllib.parse.quote(tag))
+        self.client.put(url)
+
+    def update_tags(self, project, tags):
+        """Update tag list of a project.
+
+        Replaces current tag list with list specified in tags parameter.
+
+        :param project: project to update.
+        :param tags: list of str tag names to add to the project
+
+        :returns: list of tags
+
+        """
+        url = "/projects/%s/tags" % base.getid(project)
+        for tag in tags:
+            tag = urllib.parse.quote(tag)
+        resp, body = self.client.put(url, body={"tags": tags})
+        return body['tags']
+
+    def delete_tag(self, project, tag):
+        """Remove tag from project.
+
+        :param projectd: project to remove tag from.
+        :param tag: str name of tag to remove from project
+
+        """
+        self._delete(
+            "/projects/%s/tags/%s" % (base.getid(project),
+                                      urllib.parse.quote(tag)))
+
+    def list_tags(self, project):
+        """List tags associated with project.
+
+        :param project: project to list tags for.
+
+        :returns: list of str tag names
+
+        """
+        url = "/projects/%s/tags" % base.getid(project)
+        resp, body = self.client.get(url)
+        return self._encode_tags(body['tags'])
+
+    def check_tag(self, project, tag):
+        """Check if tag is associated with project.
+
+        :param project: project to check tags for.
+        :param tag: str name of tag
+
+        :returns: true if tag is associated, false otherwise
+
+        """
+        url = "/projects/%s/tags/%s" % (base.getid(project),
+                                        urllib.parse.quote(tag))
+        try:
+            self.client.head(url)
+            # no errors means found the tag
+            return True
+        except exceptions.NotFound:
+            # 404 means tag not in project
+            return False
